(* Authors: Cláudia Nalon & Maria Júlia Dias Lima *)

Require Import ModalLogic.syntax.
Require Import ModalLogic.semantics.

Require Import FunInd.
Require Import Recdef.
Require Import Psatz.

Require Import Coq.Arith.Lt.
Require Import Coq.Logic.Classical_Prop.
Require Import Coq.Logic.Classical_Pred_Type.

Section NNF.

  (** The negation normal form takes the usual form: only constants, literals (i.e. propositions or their negations), conjunctions, disjunctions and modal operators are allowed. Note that negatins can only be applied to propositional symbols. The following function returns whether a formula is in NNF. *)
  
Fixpoint is_NNF (f:formula): Prop :=
  match f with 
  | PropsF p => True
  | ConstF ff => True
  | ConstF tt => True
  | Not (PropsF p) => True
  | Not _  => False
  | And f g => (is_NNF f ) -> (is_NNF g)
  | Or f g => (is_NNF f ) -> (is_NNF g)
  | Imp _ _ => False
  | Box f => (is_NNF f)
  | Diamond f => (is_NNF f)
  end.

(** The next definition takes a formula and returns its NNF. This is defined as a general function, because the recursion is not applied to subformulae. For instance, when taking the transformation of an implication, one of the resulting disjuncts is the negation of its antecedent, which is not a subformula of the implication. Thus, we cannot use structural induction when writing proofs about the NNF a formula. However, the recursion (and, later, our proofs) takes smaller arguments than the formulae it is applied. We take, therefore, the size of a formula as a measure for termination. *)

Function NNF (f:formula) {measure size f}: formula :=
  match f with
  | PropsF p => PropsF p
  | ConstF ff => ConstF ff
  | ConstF tt => ConstF tt
  | Not (PropsF p)=> Not (PropsF p)
  | Not (ConstF ff) => ConstF tt
  | Not (ConstF tt) => ConstF ff
  | Not (Not f) => NNF f
  | Not (And f g) => Or (NNF (Not f)) (NNF (Not g))
  | Not (Or f g) => And (NNF (Not f)) (NNF (Not g))
  | Not (Imp f g) => And (NNF f) (NNF (Not g))
  | Not (Box  f) => Diamond  (NNF (Not f))
  | Not (Diamond  f) => Box  (NNF (Not f)) 
  | And f g => And (NNF f) (NNF g)
  | Or f g => Or (NNF f) (NNF g)
  | Imp f g => Or (NNF (Not f)) (NNF g)
  | Box f => Box (NNF f)
  | Diamond f => Diamond (NNF f)
  end.
Proof.
all: try (intros;simpl;lia).
intros. simpl. apply lt_n_S. destruct g; simpl; lia.
Defined.

(** The proof that the above function is indeed terminating basically relies on properties of inequalities and are automatically obtained via the procedures implemented in coq for linear arithmetic. *)

(** The first theorem shows that the transformationn into the normal form is correct, that is, that the function does produce a formula into the desired formula. *)

Theorem NNF_is_NNF (f:formula) : 
  is_NNF (NNF f).
Proof.
  functional induction (NNF f).
  all: (simpl;auto).
Qed.

(** The next theorem shows that the formulae produced by the transformation function preserves meaning, that is, the resulting formula is semantically equivalent to the original one. The proof is by induction on the NNF of a formula. It is rather long, but it is simple: it basically relies on the induction hipothesis to show that every operation over the existing operators preserve satisfiability. Note the use of classical facts whenever needed, as the metatheory for this logic is classical. *)

Theorem eq_f_NNF_f (f : formula) :
  forall (M:Model) (w:W), sat M w f <-> sat M w (NNF f).
Proof.
  functional induction (NNF f); try (simpl; split; intro; assumption).
 - simpl; split; intro.
    + apply I. 
    + intro H0. assumption.
  - simpl; split; intro.
    + apply H. apply I.
    + intro H0. assumption.
  - simpl; split; intro.
    + apply NNPP in H. apply IHf0. apply H.
    + intro. apply H0. apply IHf0. assumption.
  - simpl; split; intro. apply not_and_or in H. destruct H as [H0|H1].
    + left. apply IHf0. assumption.
    + right. apply IHf1. assumption.
    + intro H1. destruct H1 as [H2 H3]. destruct H as [H4|H5].
      * apply IHf0 in H4. simpl in H4. apply H4. assumption.
      * apply IHf1 in H3. simpl in H3. apply H3. assumption. 
  - simpl; split; intro.
    + apply not_or_and in H. destruct H. split.
      * simpl in H. apply IHf0. assumption.
      * simpl in H. apply IHf1. assumption.
    + intro H0. destruct H. destruct H0.
      * apply IHf0 in H. simpl in H. apply H. assumption.
      * apply IHf1 in H1. simpl in H. apply H1. assumption.
  - simpl. split. intro. split. 
    + apply IHf0. apply not_or_and with (P:=(~sat M w f0)) in H.
      destruct H. apply NNPP in H. assumption.
    + apply IHf1. apply not_or_and with (P:=(~sat M w f0)) in H.
      destruct H. assumption.
    + intro. destruct H. apply IHf1 in H0.
      intro. simpl in H0. apply H0. destruct H1. apply IHf0 in H. contradiction. contradiction.
  - simpl; split; intro.
    + apply not_all_ex_not in H. destruct H as [x H0].
      exists x. apply imply_to_and in H0. unfold not in H0. destruct H0 as [H1 H2]. split.
    * assumption.
    * apply IHf0. assumption.
    + destruct H. intro. destruct H. apply H0 in H. apply IHf0 in H1.
      simpl in H1. contradiction.
  - simpl; split; intro. 
    + intros w' H0. apply not_ex_all_not with (n:=w') in H.
      apply IHf0. intro H1. unfold not in H. apply not_and_or in H.
      destruct H as [H2|H3]. apply H2. assumption. apply H3. assumption.
    + intro H0. destruct H0 as [x H1]. destruct H1. apply H in H0. apply IHf0 in H0.
      apply H0. assumption.
  - simpl; split; intro; destruct H as [H0 H1]; split; 
    try (apply IHf0); try (apply IHf1); assumption.
  - simpl; split; intro; destruct H as [H0|H1].
    + left. apply IHf0. assumption.
    + right. apply IHf1. assumption.
    + left. apply IHf0. assumption.
    + right. apply IHf1. assumption.
  - simpl; split; intro.
    + destruct H.
      * left. apply IHf0. simpl. assumption.
      * right. apply IHf1. assumption.
    + destruct H.
      * left. apply IHf0 in H. simpl in H. assumption.
      * right. apply IHf1 in H. assumption.
  - simpl; split; intro; intros w' H1; apply IHf0; apply H; assumption. 
  - simpl; split; intro; destruct H as [x H0]; exists x; destruct H0 as [H1 H2]; split;
    try assumption; try (apply IHf0; assumption).

Qed.

End NNF.

