(* Authors: Cláudia Nalon & Maria Júlia Dias Lima *)

Require Import ModalLogic.syntax.
Require Import Relations.Relation_Definitions.
Require Import Coq.Logic.Classical_Prop.


Section semantics.

  (** The semantics of the monomodal logic K is based on Kripke Models, where we have a non-empty set of worlds: *)
  
Definition W := Type.

(** a relation over the set of worlds: *)

Definition R := relation W.

(** from which a frame, there is an ordered pair of worlds and a relation, can be defined: *)

Definition Frame := (W * R) % type.

(** The evaluation function assigns a Boolean to a world and a propositional symbol: *)

Definition pi := W -> Props -> bool.

(** A Kripe model (or structure) is a frame equiped with a valuation function: *)

Definition Model := (Frame * pi) %type.

(** The satisfiability relation is recursively defined as follows: *)

Fixpoint sat (M:Model) (w:W) (f:formula): Prop :=
 let R := snd(fst(M)) in 
 let pi:= snd M in
  match f with 
  | ConstF tt => True
  | ConstF ff => False
  | PropsF p => if (pi w p) then True else False
  | Not f => ~(sat M w f)
  | And f g => (sat M w f) /\ (sat M w g)
  | Or f g => (sat M w f) \/ (sat M w g)
  | Imp f g => ~(sat M w f) \/ (sat M w g)
  | Box f => forall w':W, (R w w') -> (sat M w' f)
  | Diamond f => exists w':W, (R  w w') /\ (sat M w' f)
  end.

(** Note in the above definition that the metatheoretical connectives are classical. *)

(** The next definition says that a formula is locally satisfiable if there is a model and world that satisfies the formula according to the satisfaction relation just defined above. *)

Definition local_sat (f:formula): Prop := exists M:Model, exists w:W, sat M w f.

(** Two formula are semantically equivalent if the satisfaction relation coincide for every model and world: *)

Definition sem_equiv (f g : formula) : Prop :=
  forall (M:Model) (w:W), sat M w f = sat M w g.

End semantics.
